package org.netkernel.lang.dpml.endpoint;

import org.netkernel.lang.dpml.ast.IVariable;
import org.netkernel.layer0.nkf.INKFRequest;
import org.netkernel.layer0.nkf.INKFRequestContext;
import org.netkernel.layer0.nkf.INKFRequestReadOnly;
import org.netkernel.layer0.nkf.INKFResolutionContext;
import org.netkernel.layer0.nkf.impl.NKFSpaceImpl;
import org.netkernel.request.IRequest;
import org.netkernel.request.IRequestScopeLevel;
import org.netkernel.urii.ClassLoaderWithExports;
import org.netkernel.urii.IEndpoint;
import org.netkernel.urii.ISpace;

import java.util.ArrayList;
import java.util.HashMap;
import java.util.Iterator;
import java.util.List;
import java.util.Map;


/** A Context that provides access to any variable within a given scope
 * 
 * @author Tony Butterfield
 *
 */
public class DPMLSubrequestContext extends NKFSpaceImpl
{
	private Map mArguments;
	private IRequestScopeLevel mContextForArguments;
	private ISpace mDPMLContext;
	private String mCWU;
	private int mHash;
	private ISpace[] mCFASpaces;
	
	private static final IEndpoint sDPMLSubrequestEndpoint = new DPMLSubrequestEndpoint();
	
	public static final String ARG_SCHEME="arg:";
	public static final String ARG_SCHEME_ESC="dpml-arg:";
	
	public DPMLSubrequestContext(IRequestScopeLevel aContextForArguments, ISpace aDPMLContext, String aCWU)
	{	mArguments=new HashMap();
		mContextForArguments=aContextForArguments;
		mDPMLContext=aDPMLContext;
		mCWU=aCWU;
		
		IRequestScopeLevel s2=mContextForArguments;
		mCFASpaces = new ISpace[s2.getAvailableDepth()];
		
		for (int i=mCFASpaces.length-1; i>=0; i--)
		{	mCFASpaces[i]=s2.getSpace();
			s2=s2.getParent();
		}
		
	}
	
	public ISpace getDPMLContext()
	{	return mDPMLContext;
	}

	public void putArgument(String aName, IVariable aVariable)
	{	mArguments.put(aName,aVariable);
		mHash^=aName.hashCode();
	}
	
	public IVariable resolveName(String aName)
	{	return (IVariable)mArguments.get(aName);
	}
	
	public IRequestScopeLevel getContextForArguments()
	{	return mContextForArguments;
	}
	
	public void setContextForArguments(IRequestScopeLevel aContextForArguments)
	{	mContextForArguments=aContextForArguments;
	}

	public void handleRequestFromEndpoint(INKFRequestContext aContext) throws Exception
	{	
		int verb=aContext.getThisRequest().getVerb();
		if (verb==INKFRequestReadOnly.VERB_EXISTS)
		{	String arg=aContext.getThisRequest().getIdentifier().substring(ARG_SCHEME_ESC.length());
			aContext.createResponseFrom(mArguments.containsKey(arg));
		}
		else if (verb==INKFRequestReadOnly.VERB_SOURCE)
		{	aContext.setCWU(mCWU);
			String id=aContext.getThisRequest().getIdentifier();
			id=id.substring(ARG_SCHEME_ESC.length());
			IVariable var=resolveName(id);
			DPMLRuntime.processRequest3(aContext,var,mContextForArguments,null);
			
			// look for first intersection of the argument scope and incoming scope
			// this is our calculated high-water mark
			// (sep09-tab this may be a little to eager but at least it's safe)
			int depth1=-1;
			IRequestScopeLevel s1=aContext.getKernelContext().getRequestScope();
			while (s1!=null)
			{	int d=s1.getAvailableDepth()-1;
				if (d<mCFASpaces.length)
				{	if (s1.getSpace()==mCFASpaces[d])
					{	depth1=s1.getDepth()+1;
						break;
					}
				}
				s1=s1.getParent();
			}
			if (depth1==-1)
			{	depth1=mCFASpaces.length;
			}
			aContext.getKernelContext().declareScopeDependency(depth1);			
		}
	}
	
	public String toString()
	{
		StringBuffer sb=new StringBuffer(64);
		sb.append("DPMLSubRequestContext(");
		for (Iterator i=mArguments.keySet().iterator(); i.hasNext(); )
		{	String argName=(String)i.next();
			sb.append(argName);
			if (i.hasNext())
			{	sb.append(", ");
			}
		}
		
		sb.append(")");
		return sb.toString();
	}
	
	public boolean equals(Object aOther)
	{	boolean result=aOther==this;
		if (!result)
		{	result=aOther instanceof DPMLSubrequestContext;
			if (result)
			{	DPMLSubrequestContext other = (DPMLSubrequestContext)aOther;
				if (other.mDPMLContext==mDPMLContext && mArguments.size()==other.mArguments.size())
				{	Iterator i=mArguments.entrySet().iterator();
					Iterator j=other.mArguments.entrySet().iterator();
					while(result && i.hasNext())
					{	Map.Entry e1 = (Map.Entry)i.next();
						Map.Entry e2 = (Map.Entry)j.next();
						result = e1.getKey().equals(e2.getKey()) && e1.getValue().equals(e2.getValue());
					}
				}
				else
				{	result=false;
				}
			}
		}
		return result;
	}
	
	public int hashCode()
	{	return mHash;
	}

	
	////////////////////////////////
	// ISpace interface
	////////////////////////////////
	
	public boolean resolveEquals(Object aOther)
	{	boolean result=aOther instanceof DPMLSubrequestContext;
		if (result)
		{	DPMLSubrequestContext other = (DPMLSubrequestContext)aOther;
			if (other.mDPMLContext==mDPMLContext && mArguments.size()==other.mArguments.size())
			{	Iterator i=mArguments.entrySet().iterator();
				Iterator j=other.mArguments.entrySet().iterator();
				while(result && i.hasNext())
				{	Map.Entry e1 = (Map.Entry)i.next();
					Map.Entry e2 = (Map.Entry)j.next();
					result = e1.getKey().equals(e2.getKey());
				}
			}
			else
			{	result=false;
			}
		}
		return result;
	}
	
	////////////////////////////////
	// NKFSpaceImpl interface
	////////////////////////////////
	public void onResolve(INKFResolutionContext aContext) throws Exception
	{
		INKFRequestReadOnly toResolve = aContext.getRequestToResolve();
		if (toResolve.getVerb()==IRequest.VERB_SOURCE || toResolve.getVerb()==IRequest.VERB_EXISTS)
		{
			String identifier=toResolve.getIdentifier();
			if (identifier.startsWith(ARG_SCHEME_ESC))
			{	String name=identifier.substring(ARG_SCHEME_ESC.length());
				if (mArguments.containsKey(name))
				{	sDPMLSubrequestEndpoint.onCommissionEndpoint(getKernel(), null);
					aContext.createResolutionResponse(sDPMLSubrequestEndpoint);
				}
				else
				{	aContext.createResolutionTerminateResponse();
				}
			}
		}		
	}	

}
